////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   :  "     "
//  * Details   : http://alexander-bagel.blogspot.ru/2013/05/intercept2.html
//  * Unit Name : SpliceHelper
//  * Purpose   :     HotPatch
//  * Author    :  (Rouse_) 
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.com/
//  ****************************************************************************
//

unit SpliceHelper;

interface

uses
  Windows;

const
  LOCK_JMP_OPKODE: Word = $F9EB;
  JMP_OPKODE: Word = $E9;

type
  //      JMP NEAR OFFSET
  TNearJmpSpliceRec = packed record
    JmpOpcode: Byte;
    Offset: DWORD;
  end;
  
  THotPachSpliceData = packed record
    FuncAddr: FARPROC;
    SpliceRec: TNearJmpSpliceRec;
    LockJmp: Word;
  end;

const
  NearJmpSpliceRecSize = SizeOf(TNearJmpSpliceRec);
  LockJmpOpcodeSize = SizeOf(Word);

  procedure SpliceNearJmp(FuncAddr: Pointer; NewData: TNearJmpSpliceRec);
  procedure SpliceLockJmp(FuncAddr: Pointer; NewData: Word);
  procedure InitHotPatchSpliceRec(const LibraryName, FunctionName: string;
    InterceptHandler: Pointer; out HotPathSpliceRec: THotPachSpliceData);
  procedure InitHotPatchSpliceRecEx(const LibraryName, FunctionName: string;
    InterceptHandler, Trampoline: Pointer; out HotPathSpliceRec: THotPachSpliceData);

implementation

//        
procedure SpliceNearJmp(FuncAddr: Pointer; NewData: TNearJmpSpliceRec);
var
  OldProtect: DWORD;
begin
  VirtualProtect(FuncAddr, NearJmpSpliceRecSize,
    PAGE_EXECUTE_READWRITE, OldProtect);
  try
    Move(NewData, FuncAddr^, NearJmpSpliceRecSize);
  finally
    VirtualProtect(FuncAddr, NearJmpSpliceRecSize,
      OldProtect, OldProtect);
  end;
end;

//        
procedure SpliceLockJmp(FuncAddr: Pointer; NewData: Word);
var
  OldProtect: DWORD;
begin
  VirtualProtect(FuncAddr, LockJmpOpcodeSize, PAGE_EXECUTE_READWRITE, OldProtect);
  try
    asm
      mov  ax, NewData
      mov  ecx, FuncAddr
      lock xchg word ptr [ecx], ax
    end;
  finally
    VirtualProtect(FuncAddr, LockJmpOpcodeSize, OldProtect, OldProtect);
  end;
end;

//      
procedure InitHotPatchSpliceRec(const LibraryName, FunctionName: string;
  InterceptHandler: Pointer; out HotPathSpliceRec: THotPachSpliceData);
begin
  //     
  HotPathSpliceRec.FuncAddr :=
    GetProcAddress(GetModuleHandle(PChar(LibraryName)), PChar(FunctionName));
  //      ,    
  Move(HotPathSpliceRec.FuncAddr^, HotPathSpliceRec.LockJmp, LockJmpOpcodeSize);
  //   JMP NEAR
  HotPathSpliceRec.SpliceRec.JmpOpcode := JMP_OPKODE;
  //    (  NearJmpSpliceRecSize  ,
  // ..     )
  HotPathSpliceRec.SpliceRec.Offset :=
    PAnsiChar(InterceptHandler) - PAnsiChar(HotPathSpliceRec.FuncAddr);
end;

//           
procedure InitHotPatchSpliceRecEx(const LibraryName, FunctionName: string;
  InterceptHandler, Trampoline: Pointer; out HotPathSpliceRec: THotPachSpliceData);
var
  OldProtect: DWORD;
  TrampolineSplice: TNearJmpSpliceRec;
begin
  //     
  HotPathSpliceRec.FuncAddr :=
    GetProcAddress(GetModuleHandle(PChar(LibraryName)), PChar(FunctionName));
  //      ,    
  Move(HotPathSpliceRec.FuncAddr^, HotPathSpliceRec.LockJmp, LockJmpOpcodeSize);

  //  
  VirtualProtect(Trampoline, LockJmpOpcodeSize + NearJmpSpliceRecSize,
    PAGE_EXECUTE_READWRITE, OldProtect);
  try
    Move(HotPathSpliceRec.LockJmp, Trampoline^, LockJmpOpcodeSize);
    TrampolineSplice.JmpOpcode := JMP_OPKODE;
    TrampolineSplice.Offset := PAnsiChar(HotPathSpliceRec.FuncAddr) -
      PAnsiChar(Trampoline) - NearJmpSpliceRecSize;
    Trampoline := PAnsiChar(Trampoline) + LockJmpOpcodeSize;
    Move(TrampolineSplice, Trampoline^, SizeOf(TNearJmpSpliceRec));
  finally
    VirtualProtect(Trampoline, LockJmpOpcodeSize + NearJmpSpliceRecSize,
      OldProtect, OldProtect);
  end;

  //   JMP NEAR
  HotPathSpliceRec.SpliceRec.JmpOpcode := JMP_OPKODE;
  //    (  NearJmpSpliceRecSize  ,
  // ..     )
  HotPathSpliceRec.SpliceRec.Offset :=
    PAnsiChar(InterceptHandler) - PAnsiChar(HotPathSpliceRec.FuncAddr);
end;

end.
